"""Test BoundingBox class"""

import pytest

from bmi_topography import BoundingBox, Topography

VALID_LL = (Topography.DEFAULT["south"], Topography.DEFAULT["west"])
VALID_UR = (Topography.DEFAULT["north"], Topography.DEFAULT["east"])


def test_valid_bbox():
    BoundingBox(VALID_LL, VALID_UR)


def test_lower_left_not_iterable():
    with pytest.raises(ValueError):
        BoundingBox(0, VALID_UR)


def test_lower_left_not_coordinate_pair():
    with pytest.raises(ValueError):
        BoundingBox([0], VALID_UR)


def test_upper_right_not_iterable():
    with pytest.raises(ValueError):
        BoundingBox(VALID_LL, 0)


def test_upper_right_not_coordinate_pair():
    with pytest.raises(ValueError):
        BoundingBox(VALID_LL, [0])


def test_south_out_of_range():
    with pytest.raises(ValueError):
        BoundingBox((400, 0), VALID_UR)


def test_north_out_of_range():
    with pytest.raises(ValueError):
        BoundingBox(VALID_LL, (400, 0))


def test_south_greater_than_north():
    with pytest.raises(ValueError):
        BoundingBox((90, 0), (0, 0))


def test_west_out_of_range():
    with pytest.raises(ValueError):
        BoundingBox((0, 400), VALID_UR)


def test_east_out_of_range():
    with pytest.raises(ValueError):
        BoundingBox(VALID_LL, (0, 400))


def test_west_greater_than_east():
    with pytest.raises(ValueError):
        BoundingBox((0, 90), (0, 0))



"""Test config file"""

import yaml

from bmi_topography import Topography

CONFIG_FILE = "config.yaml"
DEM_TYPE = "SRTMGL3"
OUTPUT_FORMAT = "GTiff"


def test_read_config(shared_datadir):
    with open(shared_datadir / CONFIG_FILE) as fp:
        conf = yaml.safe_load(fp).get("bmi-topography", {})
    assert conf["dem_type"] == DEM_TYPE
    assert conf["output_format"] == OUTPUT_FORMAT


def test_set_default_config():
    conf = Topography.DEFAULT.copy()
    assert conf["dem_type"] == DEM_TYPE
    assert conf["output_format"] == OUTPUT_FORMAT


def test_pass_config_parameters():
    conf = Topography.DEFAULT.copy()
    topo = Topography(**conf)
    assert topo.dem_type == DEM_TYPE
    assert topo.output_format == OUTPUT_FORMAT



from collections.abc import Iterable


class BoundingBox:
    """Represent a simple latitude-longitude bounding box.

    Examples
    --------
    Create a bounding box:

    >>> from bmi_topography import BoundingBox
    >>> bbox = BoundingBox((33, 111), (44, 122))
    >>> print(bbox)
    [(33, 111), (44, 122)]

    """

    def __init__(self, lower_left: tuple[float], upper_right: tuple[float]) -> None:
        self._lower_left = lower_left
        self._upper_right = upper_right

        if not isinstance(self.lower_left, Iterable) or len(self.lower_left) != 2:
            raise ValueError(
                f"lower left coordinate ({self.lower_left}) must have two elements"
            )

        if not isinstance(self.upper_right, Iterable) or len(self.upper_right) != 2:
            raise ValueError(
                f"upper right coordinate ({self.upper_right}) must have two elements"
            )

        if self.south > 90 or self.south < -90:
            raise ValueError(f"south coordinate ({self.south}) must be in [-90,90]")

        if self.north > 90 or self.north < -90:
            raise ValueError(f"north coordinate ({self.north}) must be in [-90,90]")

        if self.south > self.north:
            raise ValueError(
                f"south coordinate ({self.south}) must be less than north ({self.north})"
            )

        if self.west > 180 or self.west < -180:
            raise ValueError(f"west coordinate ({self.west}) must be in [-180,180]")

        if self.east > 180 or self.east < -180:
            raise ValueError(f"east coordinate ({self.east}) must be in [-180,180]")

        if self.west > self.east:
            raise ValueError(
                f"west coordinate ({self.west}) must be less than east ({self.east})"
            )

    @property
    def lower_left(self):
        """The southwest corner of the box, given by tuple of *(south, west)*."""
        return self._lower_left

    @property
    def upper_right(self):
        """The northeast corner of the box, given by tuple of *(north, east)*."""
        return self._upper_right

    @property
    def south(self):
        return self.lower_left[0]

    @property
    def west(self):
        return self.lower_left[1]

    @property
    def north(self):
        return self.upper_right[0]

    @property
    def east(self):
        return self.upper_right[1]

    def __str__(self):
        s = f"[{self.lower_left}, {self.upper_right}]"
        return s



import os
import warnings
from functools import partial
from pathlib import Path

from .errors import BadApiKeySource, BadKeyError, MissingKeyError


class ApiKey:
    """Store an API key to use when fetching topography data from OpenTopography.

    Parameters
    ----------
    api_key : str
        An API key as a (non-empty) string.
    source : str, optional
        A string that indicates where the key came from. Possible
        values are: *user*, *env*, *file*, and *demo*.

    Raises
    ------
    BadKeyError
        The provided API key is invalid.
    MissingKeyError
        A key could not be found in the usual places.
    BadApiKeySource
        An invalid source was provided.

    Examples
    --------
    >>> from bmi_topography.api_key import ApiKey
    >>> api_key = ApiKey("foobar")
    >>> api_key
    ApiKey('foobar', source='user')
    """

    DEMO_API_KEY = "demoapikeyot2022"
    API_KEY_FILES = (".opentopography.txt", "~/.opentopography.txt")
    API_KEY_ENV_VAR = "OPENTOPOGRAPHY_API_KEY"

    def __init__(self, api_key, source="user"):
        if not isinstance(api_key, str) or len(api_key) == 0:
            raise BadKeyError("invalid API key")
        self._api_key = api_key
        self._source = ApiKey._validate_source(source)

        if self.is_demo_key():
            warnings.warn(
                "You are using a demo key to fetch data from OpenTopography, functionality "
                "will be limited. See https://bmi-topography.readthedocs.io/en/latest/#api-key "
                "for more information."
            )

    @staticmethod
    def _validate_source(source):
        valid_sources = ["user", "env", "file", "demo"]
        if isinstance(source, str) and source.split(":")[0] in valid_sources:
            return source
        else:
            raise BadApiKeySource(
                f"{source}: Invalid source (not one of {', '.join(valid_sources)}"
            )

    @classmethod
    def from_env(cls):
        """Get the key from and environment variables."""
        try:
            api_key = os.environ[ApiKey.API_KEY_ENV_VAR]
        except KeyError:
            raise MissingKeyError(f"unable to find key ({ApiKey.API_KEY_ENV_VAR})")
        else:
            return cls(api_key, source="env")

    @classmethod
    def from_file(cls):
        """Read the key from a file."""
        if filepath := _find_first_of(ApiKey.API_KEY_FILES):
            with open(filepath) as fp:
                api_key = fp.read().strip()
        else:
            raise MissingKeyError(
                f"unable to find key ({', '.join(ApiKey.API_KEY_FILES)})"
            )
        return cls(api_key, source=f"file:{filepath}")

    @classmethod
    def from_sources(cls, api_key=None):
        """Get a key from the first of a series of sources.

        Look for a key from the following sources, returning the first found:
        (1) provided by a user through the *api_key* keyword,
        (2) provided by an environment variable,
        (3) provided in a text file, and
        (4) use a demo key
        """
        for from_source in [partial(cls, api_key), cls.from_env, cls.from_file]:
            try:
                return from_source()
            except (BadKeyError, MissingKeyError):
                pass
        return cls(ApiKey.DEMO_API_KEY, source="demo")

    @property
    def api_key(self):
        """The API key."""
        return self._api_key

    @property
    def source(self):
        """Where the API key came from."""
        return self._source

    def is_demo_key(self):
        """Check if the key is a demo key."""
        return self._api_key == ApiKey.DEMO_API_KEY

    def __repr__(self):
        return f"ApiKey({self.api_key!r}, source={self.source!r})"

    def __str__(self):
        return self.api_key

    def __len__(self):
        return len(self.api_key)

    def __eq__(self, val):
        try:
            return self.api_key == val.api_key
        except AttributeError:
            return str(self) == val


def _find_first_of(files):
    """Find the first existing file from a list of files."""
    found = None
    for path in (Path(name) for name in files):
        if path.expanduser().is_file():
            found = path.expanduser()
            break
    return found



import os
import pathlib
import shutil
from itertools import chain

import nox

PROJECT = "bmi-topography"
PACKAGE = PROJECT.replace("-", "_")
HERE = pathlib.Path(__file__)
ROOT = HERE.parent
PATHS = [PACKAGE, "docs", "examples", "tests", HERE.name]
PYTHON_VERSIONS = ["3.10", "3.11", "3.12"]


@nox.session(python=PYTHON_VERSIONS)
def test(session: nox.Session) -> None:
    """Run the tests."""
    session.install(".[testing]")

    args = [
        "--cov",
        PACKAGE,
        "-vvv",
    ] + session.posargs

    if "CI" in os.environ:
        args.append(f"--cov-report=xml:{ROOT.absolute()!s}/coverage.xml")
        args.append(f"--cov-config={ROOT.absolute()!s}/pyproject.toml")
    session.run("pytest", *args)

    if "CI" not in os.environ:
        session.run("coverage", "report", "--ignore-errors", "--show-missing")


@nox.session(name="test-bmi", python=PYTHON_VERSIONS, venv_backend="conda")
def test_bmi(session: nox.Session) -> None:
    """Test the Basic Model Interface."""
    session.install("bmi-tester>=0.5.9")
    session.install(".")
    session.run(
        "bmi-test",
        f"{PACKAGE}:BmiTopography",
        "--config-file",
        f"{ROOT}/examples/config.yaml",
        "--root-dir",
        "examples",
        "-vvv",
    )


@nox.session(name="test-cli", python=PYTHON_VERSIONS)
def test_cli(session: nox.Session) -> None:
    """Test the command line interface."""
    session.install(".")
    session.run(PROJECT, "--version")
    session.run(PROJECT, "--help")


@nox.session(name="check-notebooks", python=PYTHON_VERSIONS[-1])
def check_notebooks(session: nox.Session) -> None:
    """Run the example notebooks."""
    session.install(".[testing,examples]")
    session.install("nbmake")

    args = [
        "examples",
        "--nbmake",
        "--nbmake-kernel=python3",
        "--nbmake-timeout=3000",
        "-vvv",
    ] + session.posargs

    session.run("pytest", *args)


@nox.session
def lint(session: nox.Session) -> None:
    """Clean lint and assert style."""
    session.install("pre-commit")
    session.run("pre-commit", "run", "--all-files")


@nox.session(name="prepare-docs")
def prepare_docs(session: nox.Session) -> None:
    """Update docs source before building."""
    session.run("sphinx-apidoc", "-f", "-o", "docs/source/api", PACKAGE)
    for file in ["README", "CHANGES", "CONTRIBUTING", "CODE-OF-CONDUCT", "LICENSE"]:
        session.run(
            "pandoc", "--to", "rst", f"{file}.md", "--output", f"docs/source/{file}.rst"
        )


@nox.session(name="build-docs", venv_backend="conda")
def build_docs(session: nox.Session) -> None:
    """Build the docs."""
    session.conda_install("--file", "docs/requirements.txt")
    session.install("-e", ".")

    prepare_docs(session)

    if os.path.exists("build"):
        shutil.rmtree("build")
    session.run("sphinx-build", "-b", "html", "-W", "docs/source", "build/html")


@nox.session
def build(session: nox.Session) -> None:
    """Build source and binary distributions."""
    session.install(".[build]")
    session.run("python", "-m", "build")


@nox.session
def release(session):
    """Tag, build, and publish a new release to PyPI."""
    session.install(".[build]")
    session.run("fullrelease")


@nox.session(name="testpypi")
def publish_testpypi(session):
    """Upload package to TestPyPI."""
    build(session)
    session.run("twine", "check", "dist/*")
    session.run(
        "twine",
        "upload",
        "--skip-existing",
        "--repository-url",
        "https://test.pypi.org/legacy/",
        "dist/*",
    )


@nox.session(name="pypi")
def publish_pypi(session):
    """Upload package to PyPI."""
    build(session)
    session.run("twine", "check", "dist/*")
    session.run(
        "twine",
        "upload",
        "--skip-existing",
        "--repository-url",
        "https://upload.pypi.org/legacy/",
        "dist/*",
    )


@nox.session(python=False)
def clean(session):
    """Remove virtual environments, build files, and caches."""
    shutil.rmtree("build", ignore_errors=True)
    shutil.rmtree("dist", ignore_errors=True)
    shutil.rmtree("docs/build", ignore_errors=True)
    shutil.rmtree(f"{PACKAGE}.egg-info", ignore_errors=True)
    shutil.rmtree(".pytest_cache", ignore_errors=True)
    shutil.rmtree(".venv", ignore_errors=True)
    if os.path.exists("coverage.xml"):
        os.remove("coverage.xml")
    if os.path.exists(".coverage"):
        os.remove(".coverage")
    for p in chain(ROOT.rglob("*.py[co]"), ROOT.rglob("__pycache__")):
        if p.is_dir():
            p.rmdir()
        else:
            p.unlink()


@nox.session(python=False)
def nuke(session):
    """Clean and also remove the .nox directory."""
    clean(session)
    shutil.rmtree(".nox", ignore_errors=True)



"""Base class to access elevation data"""

import os
import urllib
import warnings
from pathlib import Path

import requests
import rioxarray
from rasterio.crs import CRS
from rasterio.errors import CRSError

from .api_key import ApiKey
from .bbox import BoundingBox


class Topography:
    """Fetch and cache land elevation data from OpenTopography."""

    SCHEME = "https"
    NETLOC = "portal.opentopography.org"
    PATH = "/API/globaldem"

    DEFAULT = {
        "dem_type": "SRTMGL3",
        "south": 36.738884,
        "north": 38.091337,
        "west": -120.168457,
        "east": -118.465576,
        "output_format": "GTiff",
        "cache_dir": "~/.bmi_topography",
    }

    VALID_DEM_TYPES = (
        "SRTMGL3",
        "SRTMGL1",
        "SRTMGL1_E",
        "AW3D30",
        "AW3D30_E",
        "SRTM15Plus",
        "NASADEM",
        "COP30",
        "COP90",
    )
    VALID_OUTPUT_FORMATS = {"GTiff": "tif", "AAIGrid": "asc", "HFA": "img"}

    def __init__(
        self,
        dem_type=None,
        south=None,
        north=None,
        west=None,
        east=None,
        output_format=None,
        cache_dir=None,
        api_key=None,
    ):
        self._api_key = ApiKey.from_sources(api_key)
        # if api_key is None:
        #     self._api_key = find_user_api_key() or use_demo_key()
        # else:
        #     self._api_key = api_key

        if dem_type in Topography.VALID_DEM_TYPES:
            self._dem_type = dem_type
        else:
            raise ValueError(f"dem_type must be one of {Topography.VALID_DEM_TYPES}.")

        if output_format in Topography.VALID_OUTPUT_FORMATS.keys():
            self._output_format = output_format
            self._file_extension = Topography.VALID_OUTPUT_FORMATS[output_format]
        else:
            raise ValueError(
                "output_format must be one of %s."
                % [k for k in Topography.VALID_OUTPUT_FORMATS.keys()]
            )

        self._bbox = BoundingBox((south, west), (north, east))

        self._da = None

        if cache_dir is None:
            cache_dir = Path(Topography.DEFAULT["cache_dir"])
        self._cache_dir = Path(cache_dir).expanduser().resolve().absolute()

    @property
    def dem_type(self):
        return str(self._dem_type)

    @property
    def output_format(self):
        return str(self._output_format)

    @property
    def file_extension(self):
        return str(self._file_extension)

    @property
    def bbox(self):
        return self._bbox

    @property
    def cache_dir(self):
        return self._cache_dir

    @staticmethod
    def data_url():
        return urllib.parse.urlunparse(
            (Topography.SCHEME, Topography.NETLOC, Topography.PATH, "", "", "")
        )

    def fetch(self):
        """Download and locally store topography data.

        Returns:
            pathlib.Path: The path to the downloaded file
        """
        fname = Path(
            self.cache_dir
        ) / "{dem_type}_{south}_{west}_{north}_{east}.{ext}".format(
            dem_type=self.dem_type,
            south=self.bbox.south,
            north=self.bbox.north,
            west=self.bbox.west,
            east=self.bbox.east,
            ext=self.file_extension,
        )

        if not fname.is_file():
            self.cache_dir.mkdir(exist_ok=True)

            params = {
                "demtype": self.dem_type,
                "south": self.bbox.south,
                "north": self.bbox.north,
                "west": self.bbox.west,
                "east": self.bbox.east,
                "outputFormat": self.output_format,
            }
            if self._api_key:
                params["API_Key"] = str(self._api_key)

            response = requests.get(Topography.data_url(), params=params, stream=True)
            if response.status_code == 401:
                if self._api_key.source == "demo":
                    msg = os.linesep.join(
                        "It looks like you are using a demo key. This error may be the result of "
                        "you reaching your maximum number of downloads."
                    )
                else:
                    msg = (
                        "It looks like you are using a user-supplied key. This error may mean "
                        "that your key is out of date or there is a typo in the supplied key. "
                        f"(source={self._api_key.source})"
                    )
                response.reason = os.linesep.join([response.reason, "", msg, ""])
            response.raise_for_status()

            with fname.open("wb") as fp:
                for chunk in response.iter_content(chunk_size=None):
                    fp.write(chunk)

        return fname.absolute()

    @property
    def da(self):
        return self._da

    def load(self):
        """Load a cached topography data file into an xarray DataArray.

        Returns:
            xarray.DataArray: A container for the data
        """
        if self._da is None:
            self._da = rioxarray.open_rasterio(self.fetch())
            self._da.name = self.dem_type

            self._da.attrs["units"] = "unknown"
            try:
                crs = CRS.from_wkt(self._da.spatial_ref.crs_wkt)
            except (AttributeError, CRSError):
                warnings.warn(
                    "A CRS cannot be identified for these data. "
                    "Grid units will be set to 'unknown'."
                )
            else:
                if crs.is_geographic:
                    self._da.attrs["units"] = "degrees"
                else:
                    self._da.attrs["units"] = crs.linear_units

        return self._da



class BmiTopographyError(Exception):
    """Base exception for all bmi_topography errors."""

    pass


class MissingKeyError(BmiTopographyError):
    """Raise if an API key cannot be found."""

    pass


class BadKeyError(BmiTopographyError):
    """Raise for an invalid key."""

    pass


class BadApiKeySource(BmiTopographyError):
    """Raise for an invalid API key source."""

    pass
